<?php
// $Id: datasync.job.inc,v 1.1.2.19 2009/07/24 18:41:00 andrewlevine Exp $

class datasync_job {
  //class instance info
  public $transaction_status;
  public $iso_level;
  public $initialized;
  public $doing_phase;
  
  //Job info
  public $job_id;
  public $type;
  public $phase;
  public $next_phase;
  public $started;
  public $completed;
  
  //immutable (not really but you shouldnt change these)
  public $INCOMPLETE;
  public $UNSTARTED;
  public $DONE;
  
  public $config;
  public $phases;
  
  /**
   * Creates an instance of a job object. If passed a type (string) instead of an ID (int),
   * a new job will be created. 
   * 
   * @param $type_or_id
   */
  function __construct($id) {
    //constants (variables)
    //can't use constants here because they aren't inherited
    $this->set_constant('INCOMPLETE', 0);
    $this->set_constant('UNSTARTED', 0);
    $this->set_constant('ROLLBACK_FAILURE', 800);
    
    $this->phases = $this->phase_defs();
    $this->config = $this->config_defs();
    
    $this->transaction_status = array();
    $this->doing_phase = FALSE;
    $this->initialized = FALSE;
    
    //load job with an id
    if (ctype_digit($id)) {
      $this->job_id = $id;
    }
    else {
      throw new DataSyncException("Could not create job with non-number id %jid.", array('%jid' => $id), NULL, $this);
    }
  }
  
  
  /**
   * Function that makes sure job instance fields are properly populated. Must be called by
   * class methods thats uses these variables from the datasync_jobs table. Allows lazy
   * instantiation of job objects that we may not need info from.
   */
  public function construct_from_db($force = FALSE) {
    if ($force || !$this->initialized) {
      $result = db_query("SELECT * FROM {datasync_jobs} WHERE job_id=%d", $this->job_id);
      $row = db_fetch_array($result);
      if ($result !== FALSE && $row) {
        foreach ($row as $col => $val) {
          $this->$col = $val;
        }
        $this->initialized = TRUE;
        
      }
      else {
        //dont set job_id because it's invalid
        throw new DataSyncException("Could not find requested job_id !jid in job queue. DB error: !dbe", array('!jid' => $this->job_id, '!dbe' => db_error()), NULL, $this);
      }
    }
  }
  
  /**
   * Define the phases you want to run in your job
   * 
   * @return array
   *   A list of phases with numerical keys and
   *   string values
   */
  public function phase_defs() {
    throw new DataSyncException("You must define the phase_defs method in you datasync job class in order to tell datasync what you want it to do for the !type type.", array('!type' => $this->type), NULL, $this);
  }
  
  /**
   * Define configuration settings
   * 
   * @return array
   *   array of config settings
   */
  public function config_defs() {
    $config = array('phase_in_transaction' => array());
    drupal_alter('datasync_job_config', $config);
    return $config;
  }
  
  /**
   * Do the current phase ($this->phase)
   * 
   * @return boolean
   *   Always true. Will throw an exception on failure
   */
  public function do_phase() {
    datasync_debug("do_phase called");
    $this->start_transaction();
    if ($this->doing_phase) {
      throw new DataSyncException("Cannot do_phase while phase is running", array('!j' => $this->job_id,), 0, $this);
    }
    $this->doing_phase = TRUE;
    $this->construct_from_db();
    //since datasync_added is a faux phase, switch it to
    //a real phase value when initializing an object
    if ($this->phase == 'datasync_added') {
      //UPDATE table manually here instead of switch phase as to not disturb started var
      $phase_update = array(
      	'phase' => $this->next_phase_value(),
      	'job_id' => $this->job_id,
      );
      $result = drupal_write_record('datasync_jobs', $phase_update, 'job_id');
      if ($result === FALSE) {
        throw new DataSyncException("Could not switch job !j to from faux phase datasync_added to phase !s", array('!j' => $this->job_id, '!s' => $phase_update['phase']), NULL, $this);
      }
    }
    $this->construct_from_db(TRUE);
    $this->commit_transaction();
    
    $phase_uses_transaction = in_array($this->phase, $this->config['phase_in_transaction']);
    
    //call the phase function if it exists
    if (is_callable(array($this, $this->phase))) {
      if ($phase_uses_transaction) {
        $this->start_transaction();
      }
      
      //THIS ACTUALLY RUNS THE PHASE
      $result = $this->{$this->phase}();
      
      if ($phase_uses_transaction) {
        if ($result) {
          $this->commit_transaction();
        }
        else {
          $this->rollback_transaction();
        }
      }
 
      if (!$result) {
        datasync_debug('phase failed');  
        $this->fail_phase();
        throw new DataSyncException("Ran phase !p but it did not return a TRUE result. It returned !r", array('!p' => $this->phase, '!r' => $result), 0, $this);
      }
      else {
        //suceeded, $result == TRUE
        $this->complete_phase();
        module_invoke_all('datasync_phase_succeeded', $this);
      }
    }
    else {
      throw new DataSyncException("Tried to run phase !p but the proper method did not exist", array('!p' => $this->phase,), 0, $this);
    }
    
    return TRUE;
  }
  
  /**
   * Complete the current phase ($this->phase)
   * 
   * @return boolean
   *   Always true. Will throw an exception on failure
   */
  public function complete_phase() {
    datasync_debug("complete_phase called");
    $this->start_transaction();
    if (!$this->doing_phase) {
      throw new DataSyncException("Cannot complete_phase while doing_phase bit is off", array('!j' => $this->job_id,), 0, $this);
    }
    $completed = array ('completed' => time(), 'job_id' => $this->job_id);
    $result = drupal_write_record('datasync_jobs', $completed, 'job_id');
    $this->doing_phase = FALSE;
    $this->construct_from_db(TRUE);
    $this->commit_transaction();
    
    module_invoke_all('datasync_phase_completed', $this);
    return TRUE;
  }
  
  /**
   * Postpone the current phase ($this->phase)
   * 
   * @return boolean
   *   Always true. Will throw an exception on failure.
   */
  public function postpone_phase() {
    datasync_debug("postpone_phase called");
    $this->start_transaction();
    $this->construct_from_db();
    if ($this->doing_phase) {
      throw new DataSyncException("Cannot postpone_phase while phase is running", array('!j' => $this->job_id,), 0, $this);
    }
    $this->switch_next_phase($this->phase);
    $update = array ('completed' => time(), 'job_id' => $this->job_id);
    $result = drupal_write_record('datasync_jobs', $update, 'job_id');
    $this->construct_from_db(TRUE);
    $this->commit_transaction();
    
    module_invoke_all('datasync_phase_postponed', $this);
    return TRUE;
  }
  
   /**
   * Advance from the current phase ($this->phase) to the next phase which
   * is either $this->next_phase or if that is 'datasync_next', the default next phase
   * in the list
   * 
   * @return boolean
   *   Always true. Will throw an exception on failure
   */
  public function advance_phase() {
    datasync_debug("advance_phase called");
    if ($this->doing_phase) {
      throw new DataSyncException("Cannot advance_phase while another phase is running", array('!j' => $this->job_id,), 0, $this);
    }
    $this->start_transaction();
    $this->construct_from_db();
    $np = ($this->next_phase == 'datasync_next') ? $this->next_phase_value() : $this->next_phase;
    $this->switch_phase($np);
    $this->switch_next_phase(NULL);
    $this->commit_transaction();
    module_invoke_all('datasync_phase_advanced', $this);
    
    return TRUE;
  }
  
  /**
   * Take action because of a phase fail. This could mean starting from the initial phase,
   * restarting the current phase, or ending the job. Right now the default implementation only restarts the 
   * current phase.
   * 
   * @param $method
   *   Not in use yet. Defines what action on fail you want
   *   
   * @return
   *   true on success
   */
  public function fail_phase($method = NULL) {
    datasync_debug("fail_phase called on $this->job_id. method: $method");
    //right now we only restart the current phase
    $this->start_transaction();
    $this->construct_from_db();
    $this->doing_phase = FALSE;
    $this->switch_phase($this->phase);
    //refetch new data
    $this->construct_from_db(TRUE);
    $this->commit_transaction();
    
    module_invoke_all('datasync_phase_failed', $this);
    
    return TRUE;
  }
  
  /**
   * Should we proceed with the current ($this->phase) phase or postpone?
   * 
   * @return boolean
   *   By default, always returns TRUE. You can override this if your application
   *   ever needs to postpone running a phase for later.
   */
  public function phase_proceed() {
    return TRUE;
  }
  
  /**
   * Remove (delete) job
   */
  public function remove() {
    $this->start_transaction();
    db_query("DELETE FROM {datasync_job} WHERE job_id=%d", $this->job_id);
    if (db_error() || db_affected_rows() != 1) {
      throw new DataSyncException("Could not remove job !j", array('!j' => $this->job_id,), NULL, $this);
    }
    ds_variable_del('ds_extra_data_' . $this->job_id, NULL);
    $this->commit_transaction();
    module_invoke_all('datasync_job_removed', $this);
  }
  

  
  /**
   * Change the phase of a job
   * 
   * @param $phase_switch
   *   phase to change to (string)
   */
  protected function switch_phase($phase_switch) {
    datasync_debug("switch_phase called");
    $phase_update = array(
      'started' => $this->UNSTARTED,
      'phase' => $phase_switch,
      'job_id' => $this->job_id,
    );
    
    $this->start_transaction();
    $result = drupal_write_record('datasync_jobs', $phase_update, 'job_id');
    if ($result === FALSE) {
      throw new DataSyncException("Could not switch job !j to phase !s", array('!j' => $this->job_id, '!s' => $phase_switch), NULL, $this);
    }
    $this->construct_from_db(TRUE);
    $this->commit_transaction();
  }
  
  /**
   * Change the next phase of the job
   * 
   * @param $phase
   *   next phase to change to.
   */
  protected function switch_next_phase($nphase_switch = NULL) {
    datasync_debug("switch_next_phase called");
    if ($nphase_switch == NULL) {
      $nphase_switch = 'datasync_next';
    }
    $nphase_update = array(
      'next_phase' => $nphase_switch,
      'job_id' => $this->job_id,
    );
    $this->start_transaction();
    $result = drupal_write_record('datasync_jobs', $nphase_update, 'job_id');
    if ($result === FALSE) {
      throw new DataSyncException("Could not switch next phase of job !j to phase !s", array('!j' => $this->job_id, '!s' => serialize($nphase_switch)), NULL, $this);
    }
    $this->construct_from_db(TRUE);
    $this->commit_transaction();
  }
  
  /**
   * Returns the next phase after current assuming
   * we are following order of the phases array
   * (this ignores the next_phase variable).
   * 
   * @return 
   *   the key of the next phase
   */
  function next_phase_value() {
    $this->construct_from_db();
    if (!count($this->phases)) {
      throw new DataSyncException("Phases were assumed when none were set.", array(), 0, $this);
    }
    $tmp = $this->phases;
    ksort($tmp);
    //if the phase is added, 
    if ($this->phase == 'datasync_added') {
      return array_shift($tmp);
    }
    $next = FALSE;
    foreach ($tmp as $key => $phase) {
      if ($next) {
        return $phase;
      }
      if ($this->phase == $phase) {
        $next = TRUE;
      }
    }
    return 'datasync_done';
  }

  /* STATIC METHODS */
  
  /**
   * Take jobs from the job queue to start processing. Transactions make sure that nobody else
   * is working on these taken jobs and they won't be worked on until the next phase or it
   * times out
   * 
   * @param $num_jobs
   *   Number of jobs to take for processing.
   * @param $types_phases
   *   An array of types and phases to take. If you want to take a certain type, 
   *   just use the name of the type as the key and TRUE as the value. If you want to
   *   specify phases also, use the type as the key and an array of phases as the value.
   *   Example:
   *   array(
   *     'type1' => TRUE,
   *     'type2' => array('phase1','phase2','phase3'),
   *   );
   * @return array
   *   Array of jobs
   */
  static function take($num_jobs, $types_phases) {
    $filters = array();
    $args = array();
    foreach ($types_phases as $type => $phases) {
      $filter = 'type = ' . $type;
      if (is_array($phases) && count($phases)) {
        $filter .= " AND next_phase IN ('" . implode("',' ", $phases) . "')";
      }
    }
    
    //UNSTARTED == 0
    $query = "SELECT job_id FROM {datasync_jobs} WHERE started=0 AND phase!='datasync_done' ";
    if (count($filters)) {
      $query .= 'AND ( (' . implode(') OR (', $filters) . ' ) ) ';
    }
    $query .= "ORDER BY completed ASC, job_id ASC ";
    db_query("SET TRANSACTION ISOLATION LEVEL SERIALIZABLE");
    db_query('BEGIN;');

    $result = db_query_range($query, 0, $num_jobs);

    $jobs = array();
    $job_ids = array();
    while ($row = db_fetch_object($result)) {
      $handler = datasync_load_job_type($row->job_id);
      $jobs[] = new $handler($row->job_id);
      $job_ids[] = $row->job_id;
    }
    if (empty($jobs)) {

      db_query('ROLLBACK;');
      return array();
    }
    
    foreach ($job_ids as $jid) {
      //completed=0 is incomplete
      $result = db_query("UPDATE {datasync_jobs} SET started=%d, completed=%d WHERE job_id=%d", time(), 0, $jid);

      if ($error = db_error()) {
        db_query('ROLLBACK;');

        watchdog('datasync', 'Job !jid could not be taken (db error !dbe) so we rolled back. Arguments to static method take: %s1, %s2', array('!jid'=> $jid, '!dbe' => $error, '%s1' => $num_jobs, '%s2' => $types_phases), NULL);
        return array();
      }
    }

    db_query('COMMIT;');
    return $jobs;
  }
  
  /**
   * Adds a job of type $type to the job queue. Should always be called
   * with datasync_job_add which is in datasync.module.
   * @param $type
   *   The job type you want to run. It must be defined by a module that 
   *   defines a job type.
   * @param $data
   *   Any extra data for this job. It will be serialized and saved for later.
   *   
   * @return
   *   returns TRUE if job is added, false otherwise
   */
  static function add($type, $data = array()) {
    $job_insert = array(
    	'type' => $type, 
    	'phase' => 'datasync_added', 
    	'next_phase' => 'datasync_next', 
    	'started' => 0, // 0 == UNSTARTED 
    	'completed' => time(),
    );

    $result = drupal_write_record('datasync_jobs', $job_insert);
    if ($result === FALSE) {
      return FALSE;
    }
    if (!empty($data)) {
      ds_variable_set('ds_extra_data_' . $job_insert['job_id'], $data);
    }
    
    //note this is one of the only invokes that does not pass the job object
    module_invoke_all('datasync_job_added', ((object)$job_insert));
    
    return TRUE;
  }
  
  /* TRANSACTION HELPERS */
  
  /**
   * Retrieve extra data that was set (possibly in constructor)
   * 
   * @return 
   */
  public function extra_data() {
    $this->construct_from_db();
    return ds_variable_get('ds_extra_data_' . $this->job_id, NULL);
  }
  
  /**
   * Set extra data (overwrites previous set
   * 
   * @param $data
   *   Any data that you want to save in the database and to be
   *   associated with this job
   */
  public function set_extra_data($data) {
    ds_variable_set('ds_extra_data_' . $this->job_id, $data);
  }
  
  /**
   * Start transaction. If transaction is already started add a transaction hold to the stack.
   * 
   * @param $iso 
   *   Isolation level string (as defined by MySQL), NULL means SERIALIZABLE
   */
  public function start_transaction($iso= NULL) {
    if (empty($this->transaction_status)) {
      if ($iso == NULL) {
        $iso = 'SERIALIZABLE'; 
      }
      $result = db_query("SET TRANSACTION ISOLATION LEVEL %s", strtoupper($iso));
      if (!$result) {
        throw new DataSyncException("Could not set iso level for job !j iso level %iso", array('!j' => $this->job_id, '%iso' => $this->iso_level, ), NULL, $this);
      }
      $result = db_query('BEGIN;');
      if (!$result) {
        throw new DataSyncException("Could not start transaction for job !j iso level %iso", array('!j' => $this->job_id, '%iso' => $this->iso_level, ), NULL, $this);
      }
      $this->iso_level = strtoupper($iso);
    }
    if (!empty($iso) && $iso != $this->iso_level) {
      throw new DataSyncException("Can not change isolation level for transaction in job !j from %iso1 to %iso2", array('!j' => $this->job_id, '%iso1' => $this->iso_level, '%iso2' => $iso), NULL, $this);
    }
    $this->transaction_status[] = 1;
  }
  
  /**
   * If multiple start_transactions were called, pop one of them off the
   * transaction stack. Otherwise commit the current transaction.
   * 
   */
  public function commit_transaction() {
    if (count($this->transaction_status) == 1) {
      $result = db_query('COMMIT;');
      if (!$result) {
        throw new DataSyncException("Database error when committing: %e", array('%e' => $result), NULL, $this);
      }
      $this->transaction_status = array();
      $this->iso_level = NULL;
    }
    else {
      array_pop($this->transaction_status);
    }
  }
  
  /**
   * Rolls back transaction if it is open and reloads object
   * 
   */
  public function rollback_transaction() {
    if (count($this->transaction_status)) {
      $result = db_query("ROLLBACK;");
      if (!$result) {
        throw new DataSyncException("Database error when rolling back: %e", array('%e' => $result), $this->ROLLBACK_FAILURE, $this);
      }
      $this->transaction_status = array();
      $this->iso_level = NULL;
    }
    if ($this->job_id) {
      $this->construct_from_db(TRUE);
    }
  }
  
  /**
   * Set a constant (protected variable) only if it hasn't been set yet.
   * This is so we can do inheritance easier
   * @param $name
   *   name of the constant
   * @param $value
   *   value to set constant
   * @return 
   *   TRUE if the constant was set, FALSE if it wasnt (it was 
   *   already set)
   */
  protected function set_constant($name, $value) {
    if (!isset($this->$name)) {
      $this->$name = $value;
      return TRUE;
    }
    return FALSE;
  }

}

//THE CODE BELOW IS NECESSARY WHENEVER WE ARE HANDLING JOBS SO WE ARE INCLUDING IT HERE

class DataSyncException extends Exception {
  public function __construct($message, $variables=array(), $error_code=NULL, $job=NULL) {
    $btr = debug_backtrace();
    //this is a deadlock, probably during do_phase
    if ($job == NULL && $error_code == E_USER_WARNING) {
      //find datasync_job object in backtrace
      foreach ($btr as $context) {
        if ($context['class'] == 'datasync_job' && $context['function'] == 'do_phase') {
          //leave for rollback and fail_phase below
          $job = $context['object'];
        }
      }
    }
    
    //if job exists and we are not processing a rollback failure (error 800)
    if ($job && $error_code != 800) {
      $job->rollback_transaction();
      if ($job->doing_phase) {
        datasync_debug("Failing phase $job->phase for job $job->job_id after error occured.");
        $job->fail_phase();
      }
    }
    
    //datasync_debug(print_r($btr, TRUE));
    watchdog('datasync', $message, $variables, WATCHDOG_ERROR);
    watchdog('datasync', 'backtrace for DatasyncException: !btr', array('!btr' => serialize($btr)), WATCHDOG_ERROR);
    module_invoke_all('datasync_exception_thrown', array($message, $variables, $error_code, $job), $btr);
   
    parent::__construct($message, $error_code);
  }
  
  public function datasync_failure_type() {
    return __CLASS__ . $this->getLine();
  }
  
  public function __toString() {
    $file = $this->getFile();
    $line = $this->getLine();
    return __CLASS__ . ": [{$this->code}]: {$this->message} IN $file ON line $line \n";
  }
}

/**
 * Checks if debug mode is on, and if it is, writes $message to standard out and
 * an error log
 * @param $message
 *   debug message to write
 */
function datasync_debug($message) {
  global $datasync_debug;
  if ($datasync_debug) {
    $pretty = getmypid() . ': ' . date("m/d/y H:i:s"). ": $message\n";
    print $pretty;
    error_log($pretty, 3, 'ds_debug.log');
  }
}

/**
 * DataSync's error handler which is triggered on a query error.
 * At the moment it is used to detect deadlocks and then throw an exception.
 * @param $errno
 * @param $errstr
 * @param $errfile
 * @param $errline
 * @param $errcontext
 * @return 
 *   returns false so the other error handlers are called
 */
function datasync_error_handler($errno, $errstr, $errfile = NULL, $errline = NULL, $errcontext = NULL) {
  if ($errno == E_USER_WARNING && strstr($errstr, "\nquery:") && strstr($errstr, 'Deadlock found')) {
    throw new DataSyncException("We deadlocked with error string: !errstr", array('!errstr' => $errstr), E_USER_WARNING);
  }
  return FALSE;
}
